# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import copy
import itertools
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    SupportsIndex,
    Union,
    cast,
    overload,
)

from specfile.constants import TAG_NAMES, TAGS_WITH_ARG
from specfile.formatter import formatted
from specfile.macros import Macros
from specfile.sections import Section
from specfile.utils import split_conditional_macro_expansion

if TYPE_CHECKING:
    from specfile.specfile import Specfile


def get_tag_name_regex(name: str) -> str:
    """Contructs regex corresponding to the specified tag name."""
    regex = re.escape(name)
    if name in TAGS_WITH_ARG:
        regex += r"(?:\s*\(\s*[^\s)]*\s*\))?"
    elif name in ["source", "patch"]:
        regex += r"\d*"
    return regex


class Comment:
    """
    Class that represents a comment.

    Attributes:
        text: Text of the comment.
        prefix: Comment prefix (hash character usually surrounded by some amount of whitespace).
    """

    def __init__(self, text: str, prefix: str = "# ") -> None:
        self.text = text
        self.prefix = prefix

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Comment):
            return NotImplemented
        return self.text == other.text and self.prefix == other.prefix

    def __str__(self) -> str:
        return f"{self.prefix}{self.text}"

    @formatted
    def __repr__(self) -> str:
        return f"Comment({self.text!r}, {self.prefix!r})"


class Comments(collections.UserList):
    """
    Class that represents comments associated with a tag, that is consecutive comment lines
    located directly above a tag definition.

    Attributes:
        data: List of individual comments.
    """

    def __init__(
        self,
        data: Optional[List[Comment]] = None,
        preceding_lines: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs a `Comments` object.

        Args:
            data: List of individual comments.
            preceding_lines: Extra lines that precede comments associated with a tag.

        Returns:
            Constructed instance of `Comments` class.
        """
        super().__init__()
        if data is not None:
            self.data = data.copy()
        self._preceding_lines = (
            preceding_lines.copy() if preceding_lines is not None else []
        )

    @formatted
    def __repr__(self) -> str:
        return f"Comments({self.data!r}, {self._preceding_lines!r})"

    def __contains__(self, item: object) -> bool:
        if isinstance(item, str):
            return item in [c.text for c in self.data]
        return item in self.data

    @property
    def raw(self) -> List[str]:
        """List of comment texts"""
        return [c.text for c in self.data]

    @overload
    def __getitem__(self, i: SupportsIndex) -> Comment:
        pass

    @overload
    def __getitem__(self, i: slice) -> "Comments":
        pass

    def __getitem__(self, i):
        if isinstance(i, slice):
            return Comments(self.data[i], self._preceding_lines)
        else:
            return self.data[i]

    @overload
    def __setitem__(self, i: SupportsIndex, item: Union[Comment, str]) -> None:
        pass

    @overload
    def __setitem__(
        self, i: slice, item: Union[Iterable[Comment], Iterable[str]]
    ) -> None:
        pass

    def __setitem__(self, i, item):
        if isinstance(i, slice):
            for i0, i1 in enumerate(range(len(self.data))[i]):
                if isinstance(item[i0], str):
                    self.data[i1].text = item[i0]
                else:
                    self.data[i1] = item[i0]
        else:
            if isinstance(item, str):
                self.data[i].text = item
            else:
                self.data[i] = item

    def copy(self) -> "Comments":
        return copy.copy(self)

    def append(self, item: Union[Comment, str]) -> None:
        if isinstance(item, str):
            item = Comment(item)
        self.data.append(item)

    def insert(self, i: int, item: Union[Comment, str]) -> None:
        if isinstance(item, str):
            item = Comment(item)
        self.data.insert(i, item)

    def index(self, item: Union[Comment, str], *args: Any) -> int:
        if isinstance(item, str):
            return [c.text for c in self.data].index(item, *args)
        return self.data.index(item, *args)

    def extend(self, other: Union[Iterable[Comment], Iterable[str]]) -> None:
        for item in other:
            if isinstance(item, str):
                item = Comment(item)
            self.data.append(item)

    @classmethod
    def parse(cls, lines: List[str]) -> "Comments":
        """
        Parses list of lines into comments.

        Args:
            lines: List of lines that precede a tag definition.

        Returns:
            Constructed instance of `Comments` class.
        """
        comment_regex = re.compile(r"^(\s*#\s*)(.*)$")
        comments: List[Comment] = []
        preceding_lines: List[str] = []
        for line in reversed(lines):
            m = comment_regex.match(line)
            if not m or preceding_lines:
                preceding_lines.insert(0, line)
                continue
            comments.insert(0, Comment(*reversed(m.groups())))
        return cls(comments, preceding_lines)

    def get_raw_data(self) -> List[str]:
        return self._preceding_lines + [str(i) for i in self.data]


class Tag:
    """
    Class that represents a spec file tag.

    Attributes:
        name: Name of the tag.
        value: Literal value of the tag as stored in the spec file.
        comments: List of comments associated with the tag.
    """

    def __init__(
        self,
        name: str,
        value: str,
        separator: str,
        comments: Comments,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
        context: Optional["Specfile"] = None,
    ) -> None:
        """
        Constructs a `Tag` object.

        Args:
            name: Name of the tag.
            value: Literal value of the tag as stored in the spec file.
            expanded_value: Value of the tag after expansion by RPM.
            separator:
              Separator between name and literal value (colon usually surrounded by some
              amount of whitespace).
            comments: List of comments associated with the tag.
            prefix: Characters preceding the tag on a line.
            suffix: Characters following the tag on a line.
            context: `Specfile` instance that defines the context for macro expansions.

        Returns:
            Constructed instance of `Tag` class.
        """
        name_regexes = [
            re.compile(get_tag_name_regex(t), re.IGNORECASE) for t in TAG_NAMES
        ]
        if not name or not any(r.match(name) for r in name_regexes):
            raise ValueError(f"Invalid tag name: '{name}'")
        self.name = name
        self.value = value
        self._separator = separator
        self.comments = comments.copy()
        self._prefix = prefix or ""
        self._suffix = suffix or ""
        self._context = context

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Tag):
            return NotImplemented
        return (
            self.name == other.name
            and self.value == other.value
            and self._separator == other._separator
            and self.comments == other.comments
            and self._prefix == other._prefix
            and self._suffix == other._suffix
        )

    @formatted
    def __repr__(self) -> str:
        return (
            f"Tag({self.name!r}, {self.value!r}, {self._separator!r}, {self.comments!r}, "
            f"{self._prefix!r}, {self._suffix!r}, {self._context!r})"
        )

    def __deepcopy__(self, memo: Dict[int, Any]) -> "Tag":
        result = self.__class__.__new__(self.__class__)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == "_context":
                continue
            setattr(result, k, copy.deepcopy(v, memo))
        result._context = self._context
        return result

    @property
    def normalized_name(self) -> str:
        """
        Normalized name of the tag. The first character is capitalized
        and the rest lowercased.
        """
        return self.name.capitalize()

    @property
    def expanded_value(self) -> Optional[str]:
        """Value of the tag after expanding macros."""
        if self._context:
            return self._context.expand(self.value)
        return Macros.expand(self.value)

    def get_position(self, container: "Tags") -> int:
        """
        Gets position of this tag in a section.

        Args:
            container: `Tags` instance that contains this tag.

        Returns:
            Position expressed as line number (starting from 0).
        """
        return sum(
            len(t.comments.get_raw_data()) + 1
            for t in container[: container.index(self)]
        ) + len(self.comments.get_raw_data())


class Tags(collections.UserList):
    """
    Class that represents all tags in a certain %package section.

    Tags can be accessed by index or conveniently by name as attributes:
    ```
    # print name of the first tag
    print(tags[0].name)

    # set value of Url tag
    tags.url = 'https://example.com'

    # remove Source1 tag
    del tags.source1
    ```

    Attributes:
        data: List of individual tags.
    """

    def __init__(
        self, data: Optional[List[Tag]] = None, remainder: Optional[List[str]] = None
    ) -> None:
        """
        Constructs a `Tags` object.

        Args:
            data: List of individual tags.
            remainder: Leftover lines in a section that can't be parsed into tags.

        Returns:
            Constructed instance of `Tags` class.
        """
        super().__init__()
        if data is not None:
            self.data = data.copy()
        self._remainder = remainder.copy() if remainder is not None else []

    @formatted
    def __repr__(self) -> str:
        return f"Tags({self.data!r}, {self._remainder!r})"

    @overload
    def __getitem__(self, i: SupportsIndex) -> Tag:
        pass

    @overload
    def __getitem__(self, i: slice) -> "Tags":
        pass

    def __getitem__(self, i):
        if isinstance(i, slice):
            return Tags(self.data[i], self._remainder)
        else:
            return self.data[i]

    def __delitem__(self, i: Union[SupportsIndex, slice]) -> None:
        def delete(index):
            preceding_lines = self.data[index].comments._preceding_lines[:]
            del self.data[index]
            # preserve preceding lines of the deleted tag but compress empty lines
            if index < len(self.data):
                lines = self.data[index].comments._preceding_lines
            else:
                lines = self._remainder
            delimiter = []
            if preceding_lines and not preceding_lines[-1] or lines and not lines[0]:
                delimiter.append("")
            lines[:] = (
                list(
                    reversed(
                        list(
                            itertools.dropwhile(
                                lambda line: not line, reversed(preceding_lines)
                            )
                        )
                    )
                )
                + delimiter
                + list(itertools.dropwhile(lambda line: not line, lines))
            )

        if isinstance(i, slice):
            for index in reversed(range(len(self.data))[i]):
                delete(index)
        else:
            delete(i)

    def __contains__(self, name: object) -> bool:
        try:
            # use parent's __getattribute__() so this method can be called from __getattr__()
            data = super().__getattribute__("data")
        except AttributeError:
            return False
        return any(t.name.lower() == cast(str, name).lower() for t in data)

    def __getattr__(self, name: str) -> Tag:
        if name not in self:
            return super().__getattribute__(name)
        try:
            return self.data[self.find(name)]
        except ValueError:
            raise AttributeError(name)

    def __setattr__(self, name: str, value: Union[Tag, str]) -> None:
        if name not in self:
            return super().__setattr__(name, value)
        try:
            if isinstance(value, Tag):
                self.data[self.find(name)] = value
            else:
                self.data[self.find(name)].value = value
        except ValueError:
            raise AttributeError(name)

    def __delattr__(self, name: str) -> None:
        if name not in self:
            return super().__delattr__(name)
        try:
            del self.data[self.find(name)]
        except ValueError:
            raise AttributeError(name)

    def copy(self) -> "Tags":
        return copy.copy(self)

    def find(self, name: str) -> int:
        for i, tag in enumerate(self.data):
            if tag.name.capitalize() == name.capitalize():
                return i
        raise ValueError

    def insert(self, i: int, item: Tag) -> None:
        if i > len(self.data):
            i = len(self.data)
        if i < len(self.data):
            lines = self.data[i].comments._preceding_lines
        else:
            lines = self._remainder
        self.data.insert(i, item)
        # do not make the new tag part of a condition block (in case there is one)
        index = next(
            (i for i, line in enumerate(lines) if line.startswith("%endif")), -1
        )
        if index >= 0:
            item.comments._preceding_lines[0:0] = lines[: index + 1]
            del lines[: index + 1]

    @classmethod
    def parse(cls, section: Section, context: Optional["Specfile"] = None) -> "Tags":
        """
        Parses a section into tags.

        Args:
            section: Section to parse.
            context: `Specfile` instance that defines the context for macro expansions.

        Returns:
            Constructed instance of `Tags` class.
        """

        def regex_pattern(tag):
            name_regex = get_tag_name_regex(tag)
            return rf"^(?P<n>{name_regex})(?P<s>\s*:\s*)(?P<v>.+)"

        tag_regexes = [re.compile(regex_pattern(t), re.IGNORECASE) for t in TAG_NAMES]
        data = []
        buffer: List[str] = []
        for line in section:
            line, prefix, suffix = split_conditional_macro_expansion(line)
            # find out if there is a match for one of the tag regexes
            m = next((m for m in (r.match(line) for r in tag_regexes) if m), None)
            if m:
                data.append(
                    Tag(
                        m.group("n"),
                        m.group("v"),
                        m.group("s"),
                        Comments.parse(buffer),
                        prefix,
                        suffix,
                        context,
                    )
                )
                buffer = []
            else:
                buffer.append(prefix + line + suffix)
        return cls(data, buffer)

    def get_raw_section_data(self) -> List[str]:
        """
        Reconstructs section data from tags.

        Returns:
            List of lines forming the reconstructed section data.
        """
        result = []
        for tag in self.data:
            result.extend(tag.comments.get_raw_data())
            result.append(
                f"{tag._prefix}{tag.name}{tag._separator}{tag.value}{tag._suffix}"
            )
        result.extend(self._remainder)
        return result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import re
import string
from enum import Enum, auto
from typing import Dict, Iterable, Iterator, List, Optional, Tuple, Union, overload

from specfile.exceptions import OptionsException
from specfile.formatter import formatted


class TokenType(Enum):
    DEFAULT = auto()
    WHITESPACE = auto()
    QUOTED = auto()
    DOUBLE_QUOTED = auto()


class Token(collections.abc.Hashable):
    """
    Class that represents a single token in an option string.

    Attributes:
        type: Token type.
        value: Token value.
    """

    def __init__(self, type: TokenType, value: str) -> None:
        self.type = type
        self.value = value

    @formatted
    def __repr__(self) -> str:
        return f"Token({self.type!r}, {self.value!r})"

    def __str__(self) -> str:
        if self.type == TokenType.WHITESPACE:
            return self.value
        elif self.type == TokenType.QUOTED:
            # escape single quotes
            value = self.value.replace("'", r"\'")
            return f"'{value}'"
        elif self.type == TokenType.DOUBLE_QUOTED:
            # escape double quotes
            value = self.value.replace('"', r"\"")
            return f'"{value}"'
        # escape quotes and whitespace
        return re.sub(r"['\"\s]", r"\\\g<0>", self.value)

    def _key(self) -> tuple:
        return self.type, self.value

    def __hash__(self) -> int:
        return hash(self._key())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Token):
            return NotImplemented
        return self._key() == other._key()


class Positionals(collections.abc.MutableSequence):
    """Class that represents a sequence of positional arguments."""

    def __init__(self, options: "Options") -> None:
        """
        Constructs a `Positionals` object.

        Args:
            options: Options instance this object is tied with.

        Returns:
            Constructed instance of `Positionals` class.
        """
        self._options = options

    @formatted
    def __repr__(self) -> str:
        return f"Positionals({self._options!r})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (Positionals, list)):
            return NotImplemented
        return list(self) == other

    def __len__(self) -> int:
        return len(self._get_items())

    @overload
    def __getitem__(self, i: int) -> Union[int, str]:
        pass

    @overload
    def __getitem__(self, i: slice) -> List[Union[int, str]]:
        pass

    def __getitem__(self, i):
        items = self._get_items()
        if isinstance(i, slice):
            result = []
            for index in items[i]:
                value = self._options._tokens[index].value
                result.append(int(value) if value.isnumeric() else value)
            return result
        else:
            value = self._options._tokens[items[i]].value
            return int(value) if value.isnumeric() else value

    @overload
    def __setitem__(self, i: int, item: Union[int, str]) -> None:
        pass

    @overload
    def __setitem__(self, i: slice, item: Iterable[Union[int, str]]) -> None:
        pass

    def __setitem__(self, i, item):
        items = self._get_items()
        if isinstance(i, slice):
            for i0, i1 in enumerate(range(len(items))[i]):
                self._options._tokens[items[i1]].value = str(item[i0])
        else:
            self._options._tokens[items[i]].value = str(item)

    def __delitem__(self, i: Union[int, slice]) -> None:
        def delete(index):
            tokens = self._options._tokens
            if index == 0:
                if len(tokens) > 1 and tokens[1].type == TokenType.WHITESPACE:
                    del tokens[1]
            else:
                if tokens[index - 1].type == TokenType.WHITESPACE:
                    index -= 1
                    del tokens[index]
            del tokens[index]

        items = self._get_items()
        if isinstance(i, slice):
            for index in reversed(items[i]):
                delete(index)
        else:
            delete(items[i])

    def _get_items(self) -> List[int]:
        """
        Gets all positional arguments.

        Returns:
            List of indices of tokens that are positional arguments.
        """
        result = []
        i = 0
        while i < len(self._options._tokens):
            if self._options._tokens[i].type == TokenType.WHITESPACE:
                i += 1
                continue
            # exclude options (starting with -) and their arguments (if any)
            value = self._options._tokens[i].value
            if value.startswith("-"):
                i += 1
                if len(value) > 1:
                    if value[1] in self._options.optstring.replace(":", ""):
                        if self._options._requires_argument(value[1]):
                            if len(value) == 2:
                                if (
                                    i < len(self._options._tokens)
                                    and self._options._tokens[i].type
                                    == TokenType.WHITESPACE
                                ):
                                    i += 1
                                i += 1
                continue
            result.append(i)
            i += 1
        return result

    def insert(self, i: int, value: Union[int, str]) -> None:
        """
        Inserts a new positional argument at a specified index.

        Args:
            i: Requested index.
            value: Value of the positional argument.
        """
        items = self._get_items()
        if i > len(items):
            i = len(items)
        if items and i < len(items):
            index = items[i]
            if index > 0:
                if self._options._tokens[index - 1].type == TokenType.WHITESPACE:
                    index -= 1
        else:
            index = len(self._options._tokens)
        self._options._tokens.insert(
            index,
            Token(
                TokenType.DOUBLE_QUOTED
                if self._options._needs_quoting(value)
                else TokenType.DEFAULT,
                str(value),
            ),
        )
        if index > 0:
            self._options._tokens.insert(index, Token(TokenType.WHITESPACE, " "))


class Options(collections.abc.MutableMapping):
    """
    Class that represents macro options.

    Attributes:
        optstring: getopt-like option string containing recognized option characters.
          Option characters are ASCII letters, upper or lower-case.
          If such a character is followed by a colon, the option
          requires an argument.
        defaults: Dict specifying default arguments to options.
    """

    def __init__(
        self,
        tokens: List[Token],
        optstring: Optional[str] = None,
        defaults: Optional[Dict[str, Union[bool, int, str]]] = None,
    ) -> None:
        """
        Constructs a `Options` object.

        Args:
            tokens: List of tokens in an option string.
            optstring: String containing recognized option characters.
              Option characters are ASCII letters, upper or lower-case.
              If such a character is followed by a colon, the option
              requires an argument.
            defaults: Dict specifying default arguments to options.

        Returns:
            Constructed instance of `Options` class.
        """
        self._tokens = tokens.copy()
        self.optstring = optstring or ""
        self.defaults = defaults.copy() if defaults is not None else {}

    @formatted
    def __repr__(self) -> str:
        return f"Options({self._tokens!r}, {self.optstring!r}, {self.defaults!r})"

    def __str__(self) -> str:
        return "".join(str(t) for t in self._tokens)

    def _valid_option(self, name: str) -> bool:
        """
        Determines if a name represents a recognized option.

        Args:
            name: Name of the option.

        Returns:
            True if the option is recognized, otherwise False.
        """
        try:
            # use parent's __getattribute__() so this method can be called from __getattr__()
            optstring = super().__getattribute__("optstring")
        except AttributeError:
            return False
        return name in optstring.replace(":", "")

    def _requires_argument(self, option: str) -> bool:
        """
        Determines if an option requires an argument.

        Args:
            option: Name of the option.

        Returns:
            True if the option requires an argument, otherwise False.

        Raises:
            ValueError if the specified option is not valid.
        """
        i = self.optstring.index(option) + 1
        return i < len(self.optstring) and self.optstring[i] == ":"

    def _find_option(self, name: str) -> Tuple[Optional[int], Optional[int]]:
        """
        Searches for an option in tokens of an option string.

        Args:
            name: Name of the option.

        Returns:
            Tuple of indices where the first is the index of a token matching
            the option and the second is the index of a token matching
            its argument, or None if there is no match.
        """
        option = f"-{name}"
        for i, token in reversed(list(enumerate(self._tokens))):
            if not token.value.startswith(option):
                continue
            if token.value != option:
                return i, i
            if not self._requires_argument(name):
                return i, None
            j = i + 1
            if j == len(self._tokens):
                return i, None
            if self._tokens[j].type == TokenType.WHITESPACE:
                j += 1
            if j == len(self._tokens):
                return i, None
            if self._tokens[j].value.startswith("-"):
                return i, None
            return i, j
        return None, None

    @staticmethod
    def _needs_quoting(value):
        # if there is a whitespace, enquote the value rather than escaping it
        return any(ws in str(value) for ws in string.whitespace)

    def __getattr__(self, name: str) -> Optional[Union[bool, int, str]]:
        if not self._valid_option(name):
            return super().__getattribute__(name)
        i, j = self._find_option(name)
        if i is None:
            if self._requires_argument(name):
                return self.defaults.get(name)
            return False
        value = (
            self._tokens[j].value
            if j is not None and j > i
            else self._tokens[i].value[2:]
        )
        if not value:
            return True
        if value.isnumeric():
            return int(value)
        return value

    def __setattr__(self, name: str, value: Union[bool, int, str]) -> None:
        if not self._valid_option(name):
            return super().__setattr__(name, value)
        if self._requires_argument(name) and isinstance(value, bool):
            raise OptionsException(f"Option -{name} requires an argument.")
        if (
            not self._requires_argument(name)
            and not isinstance(value, bool)
            and value is not None
        ):
            raise OptionsException(f"Option -{name} is a flag.")
        i, j = self._find_option(name)
        if i is None:
            if value is not None and value is not False:
                if self._tokens:
                    self._tokens.append(Token(TokenType.WHITESPACE, " "))
                if value is True:
                    self._tokens.append(Token(TokenType.DEFAULT, f"-{name}"))
                elif isinstance(value, int):
                    self._tokens.append(Token(TokenType.DEFAULT, f"-{name}{value}"))
                else:
                    self._tokens.append(Token(TokenType.DEFAULT, f"-{name}"))
                    self._tokens.append(Token(TokenType.WHITESPACE, " "))
                    self._tokens.append(
                        Token(
                            TokenType.DOUBLE_QUOTED
                            if self._needs_quoting(value)
                            else TokenType.DEFAULT,
                            value,
                        )
                    )
            return
        if value is None or value is False:
            return delattr(self, name)
        if j is not None and j > i:
            if self._needs_quoting(value):
                if self._tokens[j].type not in (
                    TokenType.QUOTED,
                    TokenType.DOUBLE_QUOTED,
                ):
                    self._tokens[j].type = TokenType.DOUBLE_QUOTED
            else:
                self._tokens[j].type = TokenType.DEFAULT
            self._tokens[j].value = str(value)
        else:
            if self._needs_quoting(value):
                self._tokens[i].value = self._tokens[i].value[:2]
                self._tokens.insert(i + 1, Token(TokenType.DOUBLE_QUOTED, str(value)))
                self._tokens.insert(i + 1, Token(TokenType.WHITESPACE, " "))
            else:
                self._tokens[i].value = self._tokens[i].value[:2] + str(value)

    def __delattr__(self, name: str) -> None:
        if not self._valid_option(name):
            return super().__delattr__(name)
        i, j = self._find_option(name)
        if i is None:
            return
        if i == 0:
            if (
                j is not None
                and j + 1 < len(self._tokens)
                and self._tokens[j + 1].type == TokenType.WHITESPACE
            ):
                j += 1
        elif self._tokens[i - 1].type == TokenType.WHITESPACE:
            i -= 1
        if j is not None and j > i:
            del self._tokens[i : j + 1]
        else:
            del self._tokens[i]

    def __len__(self) -> int:
        return len(
            {
                t.value[1]
                for t in self._tokens
                if t.type != TokenType.WHITESPACE
                and t.value.startswith("-")
                and len(t.value) > 1
                and t.value[1] in self.optstring.replace(":", "")
            }
        )

    def __getitem__(self, key: str) -> Union[bool, int, str]:
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __setitem__(self, key: str, item: Union[bool, int, str]) -> None:
        try:
            return setattr(self, key, item)
        except AttributeError:
            raise KeyError(key)

    def __delitem__(self, key: str) -> None:
        try:
            return delattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __iter__(self) -> Iterator[str]:
        for option in self.optstring.replace(":", ""):
            i, _ = self._find_option(option)
            if i is not None:
                yield option

    @property
    def positional(self) -> Positionals:
        """Sequence of positional arguments."""
        return Positionals(self)

    @positional.setter
    def positional(self, value: List[Union[int, str]]) -> None:
        positionals = Positionals(self)
        positionals.clear()
        positionals.extend(value)

    @staticmethod
    def tokenize(option_string: str) -> List[Token]:
        """
        Tokenizes an option string.

        Follows the same rules as poptParseArgvString() that is used by RPM.

        Args:
            option_string: Option string.

        Returns:
            List of tokens.

        Raises:
            OptionsException if the option string is untokenizable.
        """
        result = []
        token = ""
        quote = None
        inp = list(option_string)
        while inp:
            c = inp.pop(0)
            if c == quote:
                if token:
                    result.append(
                        Token(
                            TokenType.QUOTED
                            if quote == "'"
                            else TokenType.DOUBLE_QUOTED,
                            token,
                        )
                    )
                    token = ""
                quote = None
                continue
            if quote:
                if c == "\\":
                    if not inp:
                        raise OptionsException("No escaped character")
                    c = inp.pop(0)
                    if c != quote:
                        token += "\\"
                token += c
                continue
            if c.isspace():
                if token:
                    result.append(Token(TokenType.DEFAULT, token))
                    token = ""
                whitespace = c
                while inp:
                    c = inp.pop(0)
                    if not c.isspace():
                        break
                    whitespace += c
                else:
                    result.append(Token(TokenType.WHITESPACE, whitespace))
                    break
                inp.insert(0, c)
                result.append(Token(TokenType.WHITESPACE, whitespace))
                continue
            if c in ('"', "'"):
                if token:
                    result.append(Token(TokenType.DEFAULT, token))
                    token = ""
                quote = c
                continue
            if c == "\\":
                if not inp:
                    raise OptionsException("No escaped character")
                c = inp.pop(0)
            token += c
        if quote:
            raise OptionsException("No closing quotation")
        if token:
            result.append(Token(TokenType.DEFAULT, token))
        return result



#!/usr/bin/python

import subprocess
from pathlib import Path

import fmf

# Set discover of specfile tests to a fixed commit
tree_root = Path.cwd().absolute()
tree = fmf.Tree(tree_root)
main_node = tree.find("/plans")
with main_node as data:
    data["discover"]["url"] = "https://github.com/packit/specfile.git"
    data["discover"]["ref"] = (
        subprocess.check_output(["git", "rev-parse", "HEAD"]).decode().strip()
    )


# Set discover of packit integration tests to the current main
packit_node = tree.find("/plans/packit-integration")
with packit_node as data:
    data["discover"]["ref"] = (
        subprocess.check_output(
            ["git", "ls-remote", "https://github.com/packit/packit", "main"]
        )
        .decode()
        .strip()
        .split()[0]
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
A library for parsing and manipulating RPM spec files
"""

from importlib.metadata import PackageNotFoundError, distribution

from specfile.specfile import Specfile

try:
    __version__ = distribution(__name__).version
except PackageNotFoundError:
    # package is not installed
    pass

__all__ = [
    Specfile.__name__,
]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from pathlib import Path

TESTS_DIR = Path(__file__).parent
DATA_DIR = TESTS_DIR / "data"
SPEC_MINIMAL = DATA_DIR / "spec_minimal"
SPEC_RPMAUTOSPEC = DATA_DIR / "spec_rpmautospec"
SPEC_TRADITIONAL = DATA_DIR / "spec_traditional"
SPEC_AUTOSETUP = DATA_DIR / "spec_autosetup"
SPEC_AUTOPATCH = DATA_DIR / "spec_autopatch"
SPEC_PATCHLIST = DATA_DIR / "spec_patchlist"
SPEC_INCLUDES = DATA_DIR / "spec_includes"
SPEC_MACROS = DATA_DIR / "spec_macros"
SPEC_MULTIPLE_SOURCES = DATA_DIR / "spec_multiple_sources"
SPEC_COMMENTED_PATCHES = DATA_DIR / "spec_commented_patches"
SPEC_SHELL_EXPANSIONS = DATA_DIR / "spec_shell_expansions"

SPECFILE = "test.spec"



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import copy
import re
from typing import List, Optional, SupportsIndex, Tuple, Union, overload

from specfile.formatter import formatted


class MacroDefinition:
    def __init__(
        self,
        name: str,
        body: str,
        is_global: bool,
        whitespace: Tuple[str, str, str, str],
        preceding_lines: Optional[List[str]] = None,
    ) -> None:
        self.name = name
        self.body = body
        self.is_global = is_global
        self._whitespace = whitespace
        self._preceding_lines = (
            preceding_lines.copy() if preceding_lines is not None else []
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MacroDefinition):
            return NotImplemented
        return (
            self.name == other.name
            and self.body == other.body
            and self.is_global == other.is_global
            and self._whitespace == other._whitespace
            and self._preceding_lines == other._preceding_lines
        )

    @formatted
    def __repr__(self) -> str:
        return (
            f"MacroDefinition({self.name!r}, {self.body!r}, {self.is_global!r}, "
            f"{self._whitespace!r}, {self._preceding_lines!r})"
        )

    def __str__(self) -> str:
        ws = self._whitespace
        macro = "%global" if self.is_global else "%define"
        return f"{ws[0]}{macro}{ws[1]}{self.name}{ws[2]}{self.body}{ws[3]}"

    def get_position(self, container: "MacroDefinitions") -> int:
        """
        Gets position of this macro definition in the spec file.

        Args:
            container: `MacroDefinitions` instance that contains this macro definition.

        Returns:
            Position expressed as line number (starting from 0).
        """
        return sum(
            len(md.get_raw_data()) for md in container[: container.index(self)]
        ) + len(self._preceding_lines)

    def get_raw_data(self) -> List[str]:
        result = self._preceding_lines.copy()
        ws = self._whitespace
        macro = "%global" if self.is_global else "%define"
        body = self.body.splitlines()
        if body:
            body[-1] += ws[3]
        else:
            body = [ws[3]]
        result.append(f"{ws[0]}{macro}{ws[1]}{self.name}{ws[2]}{body[0]}")
        result.extend(body[1:])
        return result


class MacroDefinitions(collections.UserList):
    """
    Class that represents all macro definitions.

    Attributes:
        data: List of individual macro definitions.
    """

    def __init__(
        self,
        data: Optional[List[MacroDefinition]] = None,
        remainder: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs a `MacroDefinitions` object.

        Args:
            data: List of individual macro definitions.
            remainder: Leftover lines that can't be parsed into macro definitions.

        Returns:
            Constructed instance of `MacroDefinitions` class.
        """
        super().__init__()
        if data is not None:
            self.data = data.copy()
        self._remainder = remainder.copy() if remainder is not None else []

    @formatted
    def __repr__(self) -> str:
        return f"MacroDefinitions({self.data!r}, {self._remainder!r})"

    def __contains__(self, name: object) -> bool:
        try:
            # use parent's __getattribute__() so this method can be called from __getattr__()
            data = super().__getattribute__("data")
        except AttributeError:
            return False
        return any(md.name == name for md in data)

    def __getattr__(self, name: str) -> MacroDefinition:
        if name not in self:
            return super().__getattribute__(name)
        try:
            return self.get(name)
        except ValueError:
            raise AttributeError(name)

    def __setattr__(self, name: str, value: Union[MacroDefinition, List[str]]) -> None:
        if name not in self:
            return super().__setattr__(name, value)
        try:
            if isinstance(value, MacroDefinition):
                self.data[self.find(name)] = value
            else:
                self.data[self.find(name)].body = value
        except ValueError:
            raise AttributeError(name)

    def __delattr__(self, name: str) -> None:
        if name not in self:
            return super().__getattribute__(name)
        try:
            del self.data[self.find(name)]
        except ValueError:
            raise AttributeError(name)

    @overload
    def __getitem__(self, i: SupportsIndex) -> MacroDefinition:
        pass

    @overload
    def __getitem__(self, i: slice) -> "MacroDefinitions":
        pass

    def __getitem__(self, i):
        if isinstance(i, slice):
            return MacroDefinitions(self.data[i], self._remainder)
        else:
            return self.data[i]

    def __delitem__(self, i: Union[SupportsIndex, slice]) -> None:
        def delete(index):
            preceding_lines = self.data[index]._preceding_lines.copy()
            del self.data[index]
            if index < len(self.data):
                self.data[index]._preceding_lines = (
                    preceding_lines + self.data[index]._preceding_lines
                )
            else:
                self._remainder = preceding_lines + self._remainder

        if isinstance(i, slice):
            for index in reversed(range(len(self.data))[i]):
                delete(index)
        else:
            delete(i)

    def copy(self) -> "MacroDefinitions":
        return copy.copy(self)

    def get(self, name: str) -> MacroDefinition:
        return self.data[self.find(name)]

    def find(self, name: str) -> int:
        for i, macro_definition in enumerate(self.data):
            if macro_definition.name == name:
                return i
        raise ValueError

    @classmethod
    def parse(cls, lines: List[str]) -> "MacroDefinitions":
        """
        Parses given lines into macro defintions.

        Args:
            lines: Lines to parse.

        Returns:
            Constructed instance of `MacroDefinitions` class.
        """

        def count_brackets(s):
            bc = pc = 0
            chars = list(s)
            while chars:
                c = chars.pop(0)
                if c == "\\" and chars:
                    chars.pop(0)
                    continue
                if c == "%" and chars:
                    c = chars.pop(0)
                    if c == "{":
                        bc += 1
                    elif c == "(":
                        pc += 1
                    continue
                if c == "{" and bc > 0:
                    bc += 1
                    continue
                if c == "}" and bc > 0:
                    bc -= 1
                    continue
                if c == "(" and pc > 0:
                    pc += 1
                    continue
                if c == ")" and pc > 0:
                    pc -= 1
                    continue
            return bc, pc

        md_regex = re.compile(
            r"""
            ^
            (\s*)                 # optional preceding whitespace
            (%(?:global|define))  # scope-defining macro definition
            (\s+)
            (\w+(?:\(.*?\))?)     # macro name with optional arguments in parentheses
            (\s+)
            (.*?)                 # macro body
            (\s*|\\)              # optional following whitespace or a backslash indicating
                                  # that the macro body continues on the next line
            $
            """,
            re.VERBOSE,
        )
        data = []
        buffer: List[str] = []
        lines = lines.copy()
        while lines:
            line = lines.pop(0)
            m = md_regex.match(line)
            if m:
                ws0, macro, ws1, name, ws2, body, ws3 = m.groups()
                if ws3 == "\\":
                    body += ws3
                    ws3 = ""
                bc, pc = count_brackets(body)
                while (bc > 0 or pc > 0 or body.endswith("\\")) and lines:
                    line = lines.pop(0)
                    body += "\n" + line
                    bc, pc = count_brackets(body)
                tokens = re.split(r"(\s+)$", body, maxsplit=1)
                if len(tokens) == 1:
                    body = tokens[0]
                else:
                    body, ws, _ = tokens
                    ws3 = ws + ws3
                data.append(
                    MacroDefinition(
                        name, body, macro == "%global", (ws0, ws1, ws2, ws3), buffer
                    )
                )
                buffer = []
            else:
                buffer.append(line)
        return cls(data, buffer)

    def get_raw_data(self) -> List[str]:
        result = []
        for macro_definition in self.data:
            result.extend(macro_definition.get_raw_data())
        result.extend(self._remainder)
        return result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import re
from enum import IntEnum
from typing import List, Optional

import rpm

from specfile.context_management import capture_stderr
from specfile.exceptions import MacroRemovalException, RPMException
from specfile.formatter import formatted

MAX_REMOVAL_RETRIES = 20


class MacroLevel(IntEnum):
    BUILTIN = -20
    DEFAULT = -15
    MACROFILES = -13
    RPMRC = -11
    CMDLINE = -7
    TARBALL = -5
    SPEC = -3
    OLDSPEC = -1
    GLOBAL = 0


class Macro(collections.abc.Hashable):
    """
    Class that represents a RPM macro.

    Attributes:
        name: Name of the macro.
        options: Options (parameters) of the macro.
        body: Macro body.
        level: Macro level (source).
        used: Indicates that the macro has been used (expanded).
    """

    def __init__(
        self,
        name: str,
        options: Optional[str],
        body: str,
        level: MacroLevel,
        used: bool,
    ) -> None:
        self.name = name
        self.options = options
        self.body = body
        self.level = level
        self.used = used

    def _key(self) -> tuple:
        return self.name, self.options, self.body, self.level, self.used

    def __hash__(self) -> int:
        return hash(self._key())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Macro):
            return NotImplemented
        return self._key() == other._key()

    @formatted
    def __repr__(self) -> str:
        return (
            f"Macro({self.name!r}, {self.options!r}, {self.body!r}, "
            f"{self.level!r}, {self.used!r})"
        )


class Macros:
    @staticmethod
    def _parse(dump: List[str]) -> List[Macro]:
        """
        Parses macros in the format of %dump output.

        Args:
            dump: List of lines in the same format as what the %dump macro outputs
                  to stderr, including newline characters.

        Returns:
            List of `Macro` instances.
        """
        # last line contains only summary
        dump.pop()
        macro_regex = re.compile(
            r"^\s*(?P<l>-?\d+)(?P<u>=|:) (?P<n>\w+)(?P<o>\(.+?\))?\t(?P<b>.*)$"
        )
        result = []
        while dump:
            line = dump.pop(0)
            # join long lines split by \
            while line.endswith("\\\n"):
                line = line[:-2] + dump.pop(0)
            # get rid of newline characters
            line = line[:-1]
            m = macro_regex.match(line)
            if m:
                result.append(
                    Macro(
                        m.group("n"),
                        m.group("o"),
                        m.group("b"),
                        MacroLevel(int(m.group("l"))),
                        m.group("u") == "=",
                    )
                )
            elif result:
                if result[-1].body:
                    result[-1].body += "\n"
                result[-1].body += line
        return result

    @classmethod
    def dump(cls) -> List[Macro]:
        """
        Dumps all macros defined in the global context.

        This is not 100% accurate, since macros can be defined multiple times,
        but only the last definition is listed.

        Returns:
            List of `Macro` instances.
        """
        with capture_stderr() as stderr:
            rpm.expandMacro("%dump")
        return cls._parse([line.decode() for line in stderr])

    @staticmethod
    def expand(expression: str) -> str:
        """
        Expands an expression in the global context.

        Args:
            expression: Expression to expand.

        Returns:
            Expanded expression.

        Raises:
            RPMException, if expansion error occurs.
        """
        try:
            with capture_stderr() as stderr:
                return rpm.expandMacro(expression)
        except rpm.error as e:
            raise RPMException(stderr=stderr) from e

    @classmethod
    def remove(cls, macro: str) -> None:
        """
        Removes all definitions of a macro in the global context.

        Args:
            macro: Macro name.

        Raises:
            MacroRemovalException, if there were too many unsuccessful
                retries to remove the macro.
        """
        # Ideally, we would loop until the macro is defined, however in rpm
        # 4.16, expanding parametrized macros may throw an exception
        # which would result in an infinite loop. Limit the number of iterations.
        retry = 0
        while retry < MAX_REMOVAL_RETRIES:
            rpm.delMacro(macro)
            try:
                if cls.expand(f"%{macro}") == f"%{macro}":
                    break
            except RPMException:
                # the macro can't be expanded, but it still exists
                retry += 1
                continue
        else:
            raise MacroRemovalException(
                f"Max attempts for removal ({MAX_REMOVAL_RETRIES}) exceeded"
            )

    @classmethod
    def define(cls, macro: str, body: str) -> None:
        """
        Defines a macro in the global context.

        Removes all existing definitions first. It is not possible to define a macro
        with options. The new macro will always have `MacroLevel.OLDSPEC` level.

        Args:
            macro: Macro name.
            body: Macro body.
        """
        cls.remove(macro)
        rpm.addMacro(macro, body)

    @classmethod
    def reinit(cls, level_threshold: MacroLevel = MacroLevel.RPMRC) -> None:
        """
        Reinitializes macros in the global context.

        Args:
            level_threshold: Only macros up to this level remain defined.
        """
        # reset everything, including macros
        rpm.reloadConfig()
        for macro in cls.dump():
            if macro.level > level_threshold:
                cls.remove(macro.name)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy

from specfile.sections import Section
from specfile.sourcelist import Sourcelist, SourcelistEntry
from specfile.tags import Comment, Comments


def test_parse():
    sourcelist = Sourcelist.parse(
        Section(
            "sourcelist",
            data=[
                "https://example.com/example-0.1.0.tar.xz",
                "",
                "# test suite",
                "tests.tar.xz",
                "",
                "",
            ],
        )
    )
    assert len(sourcelist) == 2
    assert sourcelist[0].location == "https://example.com/example-0.1.0.tar.xz"
    assert not sourcelist[0].comments
    assert sourcelist[1].location == "tests.tar.xz"
    assert sourcelist[1].comments[0].text == "test suite"


def test_get_raw_section_data():
    sourcelist = Sourcelist(
        [
            SourcelistEntry("https://example.com/example-0.1.0.tar.xz", Comments()),
            SourcelistEntry("tests.tar.xz", Comments([Comment("test suite")], [""])),
        ],
        ["", ""],
    )
    assert sourcelist.get_raw_section_data() == [
        "https://example.com/example-0.1.0.tar.xz",
        "",
        "# test suite",
        "tests.tar.xz",
        "",
        "",
    ]


def test_copy_sourcelist():
    sourcelist = Sourcelist(
        [
            SourcelistEntry("tests.tar.xz", Comments([Comment("test suite")], [""])),
        ],
    )
    shallow_copy = copy.copy(sourcelist)
    assert shallow_copy == sourcelist
    assert shallow_copy is not sourcelist
    assert shallow_copy[0] is sourcelist[0]
    deep_copy = copy.deepcopy(sourcelist)
    assert deep_copy == sourcelist
    assert deep_copy is not sourcelist
    assert deep_copy[0] is not sourcelist[0]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy
from pathlib import Path

import rpm

from specfile.spec_parser import SpecParser


def test_spec_parser_do_parse():
    parser = SpecParser(Path("."), [("dist", ".fc35")])
    spec, _ = parser._do_parse(
        (
            "Name:           test\n"
            "Version:        0.1\n"
            "Release:        1%{?dist}\n"
            "Summary:        Test package\n"
            "License:        MIT\n"
            "\n"
            "%description\n"
            "Test package\n"
        ),
    )
    assert spec.sourceHeader[rpm.RPMTAG_NAME] == "test"
    assert spec.sourceHeader[rpm.RPMTAG_VERSION] == "0.1"
    assert spec.sourceHeader[rpm.RPMTAG_RELEASE] == "1.fc35"
    assert spec.sourceHeader[rpm.RPMTAG_SUMMARY] == "Test package"
    assert spec.sourceHeader[rpm.RPMTAG_LICENSE] == "MIT"
    assert spec.prep is None


def test_copy_spec_parser():
    parser = SpecParser(Path("."), [("dist", ".fc35")])
    shallow_copy = copy.copy(parser)
    assert shallow_copy == parser
    assert shallow_copy is not parser
    deep_copy = copy.deepcopy(parser)
    assert deep_copy == parser
    assert deep_copy is not parser



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy

import pytest

from specfile.options import Options, Token, TokenType
from specfile.prep import AutosetupMacro, PatchMacro, Prep, PrepMacros, SetupMacro
from specfile.sections import Section


@pytest.mark.parametrize(
    "name, options, number",
    [
        ("%patch2", "-p1", 2),
        ("%patch0028", "-p2", 28),
        ("%patch", "-p1", 0),
        ("%patch", "-P1", 1),
        ("%patch3", "-P5", 5),
    ],
)
def test_patch_macro_number(name, options, number):
    macro = PatchMacro(
        name, Options(Options.tokenize(options), PatchMacro.OPTSTRING), " "
    )
    assert macro.number == number


def test_prep_macros_find():
    macros = PrepMacros(
        [
            SetupMacro("%setup", Options([]), ""),
            PatchMacro("%patch0", Options([]), ""),
        ]
    )
    assert macros.find("%patch0") == 1
    with pytest.raises(ValueError):
        macros.find("%autosetup")


@pytest.mark.parametrize(
    "lines_before, number, options, lines_after",
    [
        (
            ["%setup -q"],
            0,
            dict(p=1),
            ["%setup -q", "%patch0 -p1"],
        ),
        (
            ["%setup -q", "%patch0 -p1"],
            0,
            dict(p=2),
            ["%setup -q", "%patch0 -p1", "%patch0 -p2"],
        ),
        (
            ["%setup -q", "%patch999 -p1"],
            28,
            dict(p=1),
            ["%setup -q", "%patch28 -p1", "%patch999 -p1"],
        ),
        (
            ["%setup -q", "%patch999 -p1"],
            1001,
            dict(p=1),
            ["%setup -q", "%patch999 -p1", "%patch1001 -p1"],
        ),
        (
            ["%setup -q", "%{!?skip_first_patch:%patch0 -p1}", "%patch999 -p1"],
            28,
            dict(p=2, b=".patch28"),
            [
                "%setup -q",
                "%{!?skip_first_patch:%patch0 -p1}",
                "%patch28 -p2 -b .patch28",
                "%patch999 -p1",
            ],
        ),
    ],
)
def test_prep_add_patch_macro(lines_before, number, options, lines_after):
    prep = Prep.parse(Section("prep", data=lines_before))
    prep.add_patch_macro(number, **options)
    assert prep.get_raw_section_data() == lines_after


@pytest.mark.parametrize(
    "lines_before, number, lines_after",
    [
        (
            ["%setup -q", "%patch0 -p1", "%patch1 -p1", "%patch2 -p1"],
            1,
            ["%setup -q", "%patch0 -p1", "%patch2 -p1"],
        ),
        (
            ["%setup -q", "%{!?skip_first_patch:%patch0 -p1}", "%patch1 -p1"],
            0,
            ["%setup -q", "%patch1 -p1"],
        ),
    ],
)
def test_prep_remove_patch_macro(lines_before, number, lines_after):
    prep = Prep.parse(Section("prep", data=lines_before))
    prep.remove_patch_macro(number)
    assert prep.get_raw_section_data() == lines_after


def test_prep_parse():
    prep = Prep.parse(
        Section(
            "prep",
            data=[
                "%setup -q",
                "# a comment",
                "%patch0 -p1",
                "%{!?skip_patch2:%patch2 -p2}",
                "",
            ],
        )
    )
    assert prep.macros[0].name == "%setup"
    assert prep.macros[0].options.q
    assert prep.macros[1].name == "%patch0"
    assert prep.macros[1].options.p == 1
    assert prep.patch2.options.p == 2


def test_prep_get_raw_section_data():
    prep = Prep(
        PrepMacros(
            [
                SetupMacro(
                    SetupMacro.CANONICAL_NAME,
                    Options(
                        [Token(TokenType.DEFAULT, "-q")],
                        SetupMacro.OPTSTRING,
                        SetupMacro.DEFAULTS,
                    ),
                    " ",
                ),
                PatchMacro(
                    PatchMacro.CANONICAL_NAME + "0",
                    Options(
                        [Token(TokenType.DEFAULT, "-p1")],
                        PatchMacro.OPTSTRING,
                        PatchMacro.DEFAULTS,
                    ),
                    " ",
                    preceding_lines=["# a comment"],
                ),
                PatchMacro(
                    PatchMacro.CANONICAL_NAME + "2",
                    Options(
                        [Token(TokenType.DEFAULT, "-p2")],
                        PatchMacro.OPTSTRING,
                        PatchMacro.DEFAULTS,
                    ),
                    " ",
                    "%{!?skip_patch2:",
                    "}",
                ),
            ],
            [""],
        )
    )
    assert prep.get_raw_section_data() == [
        "%setup -q",
        "# a comment",
        "%patch0 -p1",
        "%{!?skip_patch2:%patch2 -p2}",
        "",
    ]


def test_copy_prep():
    prep = Prep(
        PrepMacros(
            [
                AutosetupMacro(
                    AutosetupMacro.CANONICAL_NAME,
                    Options([]),
                    "",
                ),
            ],
        )
    )
    shallow_copy = copy.copy(prep)
    assert shallow_copy == prep
    assert shallow_copy is not prep
    assert shallow_copy.macros is prep.macros
    assert shallow_copy.macros[0] is prep.macros[0]
    deep_copy = copy.deepcopy(prep)
    assert deep_copy == prep
    assert deep_copy is not prep
    assert deep_copy.macros is not prep.macros
    assert deep_copy.macros[0] is not prep.macros[0]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from typing import List


class SpecfileException(Exception):
    """Something went wrong during our execution."""


class RPMException(SpecfileException):
    """Exception related to RPM."""

    def __init__(self, stderr: List[bytes]) -> None:
        super().__init__()
        self.stderr = stderr

    def __str__(self) -> str:
        for rline in self.stderr:
            line = rline.decode().rstrip()
            if line.startswith("error:"):
                return line.split("error: ")[1]
        return "\n" + b"".join(self.stderr).decode()


class MacroRemovalException(SpecfileException):
    """Exception related to failed removal of RPM macros."""


class OptionsException(SpecfileException):
    """Exception related to processing options."""


class UnterminatedMacroException(SpecfileException):
    """Exception related to parsing unterminated macro."""


class DuplicateSourceException(SpecfileException):
    """Exception related to adding a duplicate source."""


class SourceNumberException(SpecfileException):
    """Exception related to source numbers."""



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import re
from typing import Tuple

from specfile.constants import ARCH_NAMES
from specfile.exceptions import SpecfileException, UnterminatedMacroException
from specfile.formatter import formatted
from specfile.value_parser import ConditionalMacroExpansion, ValueParser


class EVR(collections.abc.Hashable):
    """Class representing Epoch-Version-Release combination."""

    _regex = r"(?:(\d+):)?([^-]+?)(?:-([^-]+))?"

    def __init__(self, *, version: str, release: str = "", epoch: int = 0) -> None:
        self.epoch = epoch
        self.version = version
        self.release = release

    def _key(self) -> tuple:
        return self.epoch, self.version, self.release

    def __hash__(self) -> int:
        return hash(self._key())

    def __eq__(self, other: object) -> bool:
        if type(other) != self.__class__:
            return NotImplemented
        return self._key() == other._key()

    @formatted
    def __repr__(self) -> str:
        return f"EVR(epoch={self.epoch!r}, version={self.version!r}, release={self.release!r})"

    def __str__(self) -> str:
        epoch = f"{self.epoch}:" if self.epoch > 0 else ""
        release = f"-{self.release}" if self.release else ""
        return f"{epoch}{self.version}{release}"

    @classmethod
    def from_string(cls, evr: str) -> "EVR":
        m = re.match(f"^{cls._regex}$", evr)
        if not m:
            raise SpecfileException("Invalid EVR string.")
        e, v, r = m.groups()
        return cls(epoch=int(e) if e else 0, version=v, release=r or "")


class NEVR(EVR):
    """Class representing Name-Epoch-Version-Release combination."""

    _regex = r"(.+?)-" + EVR._regex

    def __init__(
        self, *, name: str, version: str, release: str = "", epoch: int = 0
    ) -> None:
        self.name = name
        super().__init__(epoch=epoch, version=version, release=release)

    def _key(self) -> tuple:
        return self.name, self.epoch, self.version, self.release

    @formatted
    def __repr__(self) -> str:
        return (
            f"NEVR(name={self.name!r}, epoch={self.epoch!r}, "
            f"version={self.version!r}, release={self.release!r})"
        )

    def __str__(self) -> str:
        return f"{self.name}-" + super().__str__()

    @classmethod
    def from_string(cls, nevr: str) -> "NEVR":
        m = re.match(f"^{cls._regex}$", nevr)
        if not m:
            raise SpecfileException("Invalid NEVR string.")
        n, e, v, r = m.groups()
        return cls(name=n, epoch=int(e) if e else 0, version=v, release=r or "")


class NEVRA(NEVR):
    """Class representing Name-Epoch-Version-Release-Arch combination."""

    _arches_regex = "(" + "|".join(re.escape(a) for a in ARCH_NAMES | {"noarch"}) + ")"
    _regex = NEVR._regex + r"\." + _arches_regex

    def __init__(
        self, *, name: str, version: str, release: str, arch: str, epoch: int = 0
    ) -> None:
        if not re.match(f"^{self._arches_regex}$", arch):
            raise SpecfileException("Invalid architecture name.")
        self.arch = arch
        super().__init__(name=name, epoch=epoch, version=version, release=release)

    def _key(self) -> tuple:
        return self.name, self.epoch, self.version, self.release, self.arch

    @formatted
    def __repr__(self) -> str:
        return (
            f"NEVRA(name={self.name!r}, epoch={self.epoch!r}, "
            f"version={self.version!r}, release={self.release!r}, "
            f"arch={self.arch!r})"
        )

    def __str__(self) -> str:
        return super().__str__() + f".{self.arch}"

    @classmethod
    def from_string(cls, nevra: str) -> "NEVRA":
        m = re.match(f"^{cls._regex}$", nevra)
        if not m:
            raise SpecfileException("Invalid NEVRA string.")
        n, e, v, r, a = m.groups()
        return cls(name=n, epoch=int(e) if e else 0, version=v, release=r, arch=a)


def get_filename_from_location(location: str) -> str:
    """
    Extracts filename from given source location.

    Follows RPM logic - target filename can be specified in URL fragment.

    Args:
        location: Location to extract filename from.

    Returns:
        Extracted filename that can be empty if there is none.
    """
    slash = location.rfind("/")
    if slash < 0:
        return location
    return location[slash + 1 :].split("=")[-1]


def split_conditional_macro_expansion(value: str) -> Tuple[str, str, str]:
    """
    Splits conditional macro expansion into its body and prefix and suffix of it.
    If the passed string isn't a conditional macro expansion, returns it as it is.

    Args:
        value: String to be split.

    Returns:
        Tuple of body, prefix, suffix. Prefix and suffix will be empty if the passed string
        isn't a conditional macro expansion.
    """
    try:
        nodes = ValueParser.parse(value)
    except UnterminatedMacroException:
        return value, "", ""
    if len(nodes) != 1:
        return value, "", ""
    node = nodes[0]
    if not isinstance(node, ConditionalMacroExpansion):
        return value, "", ""
    return "".join(str(n) for n in node.body), f"%{{{node.prefix}{node.name}:", "}"



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import copy
from typing import TYPE_CHECKING, Any, Dict, List, Optional, SupportsIndex, overload

from specfile.formatter import formatted
from specfile.macros import Macros
from specfile.sections import Section
from specfile.tags import Comments

if TYPE_CHECKING:
    from specfile.specfile import Specfile


class SourcelistEntry:
    """
    Class that represents a spec file source/patch in a %sourcelist/%patchlist.

    Attributes:
        location: Literal location of the source/patch as stored in the spec file.
        comments: List of comments associated with the source/patch.
    """

    def __init__(
        self, location: str, comments: Comments, context: Optional["Specfile"] = None
    ) -> None:
        """
        Constructs a `SourceListEntry` object.

        Args:
            location: Literal location of the source/patch as stored in the spec file.
            comments: List of comments associated with the source/patch.
            context: `Specfile` instance that defines the context for macro expansions.

        Returns:
            Constructed instance of `SourceListEntry` class.
        """
        self.location = location
        self.comments = comments.copy()
        self._context = context

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SourcelistEntry):
            return NotImplemented
        return self.location == other.location and self.comments == other.comments

    @formatted
    def __repr__(self) -> str:
        return (
            f"SourcelistEntry({self.location!r}, {self.comments!r}, {self._context!r})"
        )

    def __deepcopy__(self, memo: Dict[int, Any]) -> "SourcelistEntry":
        result = self.__class__.__new__(self.__class__)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == "_context":
                continue
            setattr(result, k, copy.deepcopy(v, memo))
        result._context = self._context
        return result

    @property
    def expanded_location(self) -> str:
        """URL of the source/patch after expanding macros."""
        if self._context:
            return self._context.expand(self.location)
        return Macros.expand(self.location)


class Sourcelist(collections.UserList):
    """
    Class that represents entries in a %sourcelist/%patchlist section.

    Attributes:
        data: List of individual sources/patches.
    """

    def __init__(
        self,
        data: Optional[List[SourcelistEntry]] = None,
        remainder: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs a `Sourcelist` object.

        Args:
            data: List of individual sources/patches.
            remainder: Leftover lines in a section that can't be parsed into sources/patches.

        Returns:
            Constructed instance of `Sourcelist` class.
        """
        super().__init__()
        if data is not None:
            self.data = data.copy()
        self._remainder = remainder.copy() if remainder is not None else []

    @formatted
    def __repr__(self) -> str:
        return f"Sourcelist({self.data!r}, {self._remainder!r})"

    @overload
    def __getitem__(self, i: SupportsIndex) -> SourcelistEntry:
        pass

    @overload
    def __getitem__(self, i: slice) -> "Sourcelist":
        pass

    def __getitem__(self, i):
        if isinstance(i, slice):
            return Sourcelist(self.data[i], self._remainder)
        else:
            return self.data[i]

    def copy(self) -> "Sourcelist":
        return copy.copy(self)

    @classmethod
    def parse(
        cls, section: Section, context: Optional["Specfile"] = None
    ) -> "Sourcelist":
        """
        Parses a section into sources/patches.

        Args:
            section: %sourcelist/%patchlist section.
            context: `Specfile` instance that defines the context for macro expansions.

        Returns:
            Constructed instance of `Sourcelist` class.
        """
        data = []
        buffer: List[str] = []
        for line in section:
            if line and not line.lstrip().startswith("#"):
                data.append(SourcelistEntry(line, Comments.parse(buffer), context))
                buffer = []
            else:
                buffer.append(line)
        return cls(data, buffer)

    def get_raw_section_data(self) -> List[str]:
        """
        Reconstructs section data from sources/patches.

        Returns:
            List of lines forming the reconstructed section data.
        """
        result = []
        for source in self.data:
            result.extend(source.comments.get_raw_data())
            result.append(source.location)
        result.extend(self._remainder)
        return result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import shutil

import pytest

from tests.constants import (
    SPEC_AUTOPATCH,
    SPEC_AUTOSETUP,
    SPEC_COMMENTED_PATCHES,
    SPEC_INCLUDES,
    SPEC_MACROS,
    SPEC_MINIMAL,
    SPEC_MULTIPLE_SOURCES,
    SPEC_PATCHLIST,
    SPEC_RPMAUTOSPEC,
    SPEC_SHELL_EXPANSIONS,
    SPEC_TRADITIONAL,
    SPECFILE,
)


@pytest.fixture(scope="function")
def spec_minimal(tmp_path):
    specfile_path = tmp_path / SPECFILE
    shutil.copyfile(SPEC_MINIMAL / SPECFILE, specfile_path)
    return specfile_path


@pytest.fixture(scope="function")
def spec_rpmautospec(tmp_path):
    specfile_path = tmp_path / SPECFILE
    shutil.copyfile(SPEC_RPMAUTOSPEC / SPECFILE, specfile_path)
    return specfile_path


@pytest.fixture(scope="function")
def spec_traditional(tmp_path):
    destination = tmp_path / "spec_traditional"
    shutil.copytree(SPEC_TRADITIONAL, destination)
    return destination / SPECFILE


@pytest.fixture(scope="function")
def spec_autosetup(tmp_path):
    destination = tmp_path / "spec_autosetup"
    shutil.copytree(SPEC_AUTOSETUP, destination)
    return destination / SPECFILE


@pytest.fixture(scope="function")
def spec_autopatch(tmp_path):
    destination = tmp_path / "spec_autopatch"
    shutil.copytree(SPEC_AUTOPATCH, destination)
    return destination / SPECFILE


@pytest.fixture(scope="function")
def spec_patchlist(tmp_path):
    destination = tmp_path / "spec_patchlist"
    shutil.copytree(SPEC_PATCHLIST, destination)
    return destination / SPECFILE


@pytest.fixture(scope="function")
def spec_includes(tmp_path):
    destination = tmp_path / "spec_includes"
    shutil.copytree(SPEC_INCLUDES, destination)
    return destination / SPECFILE


@pytest.fixture(scope="function")
def spec_macros(tmp_path):
    destination = tmp_path / "spec_macros"
    shutil.copytree(SPEC_MACROS, destination)
    return destination / SPECFILE


@pytest.fixture(scope="function")
def spec_multiple_sources(tmp_path):
    destination = tmp_path / "spec_multiple_sources"
    shutil.copytree(SPEC_MULTIPLE_SOURCES, destination)
    return destination / SPECFILE


@pytest.fixture(scope="function")
def spec_commented_patches(tmp_path):
    destination = tmp_path / "spec_commented_patches"
    shutil.copytree(SPEC_COMMENTED_PATCHES, destination)
    return destination / SPECFILE


@pytest.fixture(scope="function")
def spec_shell_expansions(tmp_path):
    destination = tmp_path / "spec_shell_expansions"
    shutil.copytree(SPEC_SHELL_EXPANSIONS, destination)
    return destination / SPECFILE



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
import re
import subprocess
import types
from dataclasses import dataclass
from pathlib import Path
from typing import Generator, List, Optional, Tuple, Type, Union, cast

import rpm

from specfile.changelog import Changelog, ChangelogEntry
from specfile.context_management import ContextManager
from specfile.exceptions import SourceNumberException, SpecfileException
from specfile.formatter import formatted
from specfile.macro_definitions import MacroDefinition, MacroDefinitions
from specfile.macros import Macro, Macros
from specfile.prep import Prep
from specfile.sections import Section, Sections
from specfile.sourcelist import Sourcelist
from specfile.sources import Patches, Sources
from specfile.spec_parser import SpecParser
from specfile.tags import Tag, Tags
from specfile.value_parser import SUBSTITUTION_GROUP_PREFIX, ValueParser


class Specfile:
    """
    Class that represents a spec file.

    Attributes:
        autosave: Whether to automatically save any changes made.
    """

    def __init__(
        self,
        path: Union[Path, str],
        sourcedir: Optional[Union[Path, str]] = None,
        autosave: bool = False,
        macros: Optional[List[Tuple[str, str]]] = None,
        force_parse: bool = False,
    ) -> None:
        """
        Constructs a `Specfile` object.

        Args:
            path: Path to the spec file.
            sourcedir: Path to sources and patches.
            autosave: Whether to automatically save any changes made.
            macros: List of extra macro definitions.
            force_parse: Whether to attempt to parse the spec file even if one or more
              sources required to be present at parsing time are not available.
              Such sources include sources referenced from shell expansions
              in tag values and sources included using the %include directive.

        Returns:
            Constructed instance of `Specfile` class.
        """
        self.autosave = autosave
        self._path = Path(path)
        self._lines = self.path.read_text().splitlines()
        self._parser = SpecParser(
            Path(sourcedir or self.path.parent), macros, force_parse
        )
        # parse here to fail early on parsing errors
        self._parser.parse(str(self))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Specfile):
            return NotImplemented
        return (
            self.autosave == other.autosave
            and self._path == other._path
            and self._lines == other._lines
            and self._parser == other._parser
        )

    @formatted
    def __repr__(self) -> str:
        return (
            f"Specfile({self.path!r}, {self._parser.sourcedir!r}, {self.autosave!r}, "
            f"{self._parser.macros!r}, {self._parser.force_parse!r})"
        )

    def __str__(self) -> str:
        return "\n".join(self._lines) + "\n"

    def __enter__(self) -> "Specfile":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[types.TracebackType],
    ) -> None:
        self.save()

    @property
    def path(self) -> Path:
        """Path to the spec file."""
        return self._path

    @path.setter
    def path(self, value: Union[Path, str]) -> None:
        self._path = Path(value)

    @property
    def sourcedir(self) -> Path:
        """Path to sources and patches."""
        return self._parser.sourcedir

    @sourcedir.setter
    def sourcedir(self, value: Union[Path, str]) -> None:
        self._parser.sourcedir = Path(value)

    @property
    def macros(self) -> List[Tuple[str, str]]:
        """List of extra macro definitions."""
        return self._parser.macros

    @property
    def force_parse(self) -> bool:
        """
        Whether to attempt to parse the spec file even if one or more
        sources required to be present at parsing time are not available.
        """
        return self._parser.force_parse

    @force_parse.setter
    def force_parse(self, value: bool) -> None:
        self._parser.force_parse = value

    @property
    def tainted(self) -> bool:
        """
        Indication that parsing of the spec file was forced and one or more
        sources required to be present at parsing time were not available
        and were replaced with dummy files.
        """
        self._parser.parse(str(self))
        return self._parser.tainted

    @property
    def rpm_spec(self) -> rpm.spec:
        """Underlying `rpm.spec` instance."""
        self._parser.parse(str(self))
        return self._parser.spec

    def reload(self) -> None:
        """Reload the spec file content."""
        self._lines = self.path.read_text().splitlines()

    def save(self) -> None:
        """Save the spec file content."""
        self.path.write_text(str(self))

    def expand(
        self,
        expression: str,
        extra_macros: Optional[List[Tuple[str, str]]] = None,
        skip_parsing: bool = False,
    ) -> str:
        """
        Expands an expression in the context of the spec file.

        Args:
            expression: Expression to expand.
            extra_macros: Extra macros to be defined before expansion is performed.
            skip_parsing: Do not parse the spec file before expansion is performed.
              Defaults to False. Mutually exclusive with extra_macros. Set this to True
              only if you are certain that the global macro context is up-to-date.

        Returns:
            Expanded expression.
        """
        if not skip_parsing:
            self._parser.parse(str(self), extra_macros)
        return Macros.expand(expression)

    def get_active_macros(self) -> List[Macro]:
        """
        Gets active macros in the context of the spec file.

        This includes built-in RPM macros, macros loaded from macro files
        and macros defined in the spec file itself.

        Returns:
            List of `Macro` objects.
        """
        self._parser.parse(str(self))
        return Macros.dump()

    @ContextManager
    def lines(self) -> Generator[List[str], None, None]:
        """
        Context manager for accessing spec file lines.

        Yields:
            Spec file lines as list of strings.
        """
        try:
            yield self._lines
        finally:
            if self.autosave:
                self.save()

    @ContextManager
    def macro_definitions(self) -> Generator[MacroDefinitions, None, None]:
        """
        Context manager for accessing macro definitions.

        Yields:
            Macro definitions in the spec file as `MacroDefinitions` object.
        """
        with self.lines() as lines:
            macro_definitions = MacroDefinitions.parse(lines)
            try:
                yield macro_definitions
            finally:
                lines[:] = macro_definitions.get_raw_data()

    @ContextManager
    def sections(self) -> Generator[Sections, None, None]:
        """
        Context manager for accessing spec file sections.

        Yields:
            Spec file sections as `Sections` object.
        """
        with self.lines() as lines:
            sections = Sections.parse(lines, context=self)
            try:
                yield sections
            finally:
                lines[:] = sections.get_raw_data()

    @property
    def parsed_sections(self) -> Sections:
        """Parsed spec file sections."""
        return Sections.parse(self.rpm_spec.parsed.splitlines())

    @ContextManager
    def tags(
        self, section: Union[str, Section] = "package"
    ) -> Generator[Tags, None, None]:
        """
        Context manager for accessing tags in a specified section.

        Args:
            section: Name of the requested section or an existing `Section` instance.
              Defaults to preamble.

        Yields:
            Tags in the section as `Tags` object.
        """
        with self.sections() as sections:
            if isinstance(section, str):
                section = cast(Section, getattr(sections, section))
            tags = Tags.parse(section, context=self)
            try:
                yield tags
            finally:
                section.data = tags.get_raw_section_data()

    @ContextManager
    def changelog(self) -> Generator[Optional[Changelog], None, None]:
        """
        Context manager for accessing changelog.

        Yields:
            Spec file changelog as `Changelog` object or None if there is no %changelog section.
        """
        with self.sections() as sections:
            try:
                section = sections.changelog
            except AttributeError:
                yield None
            else:
                changelog = Changelog.parse(section)
                try:
                    yield changelog
                finally:
                    section.data = changelog.get_raw_section_data()

    @ContextManager
    def prep(self) -> Generator[Optional[Prep], None, None]:
        """
        Context manager for accessing %prep section.

        Yields:
            Spec file %prep section as `Prep` object.
        """
        with self.sections() as sections:
            try:
                section = sections.prep
            except AttributeError:
                yield None
            else:
                prep = Prep.parse(section)
                try:
                    yield prep
                finally:
                    section.data = prep.get_raw_section_data()

    @ContextManager
    def sources(
        self,
        allow_duplicates: bool = False,
        default_to_implicit_numbering: bool = False,
        default_source_number_digits: int = 1,
    ) -> Generator[Sources, None, None]:
        """
        Context manager for accessing sources.

        Args:
            allow_duplicates: Whether to allow duplicate entries when adding new sources.
            default_to_implicit_numbering: Use implicit numbering (no source numbers) by default.
            default_source_number_digits: Default number of digits in a source number.

        Yields:
            Spec file sources as `Sources` object.
        """
        with self.sections() as sections, self.tags() as tags:
            sourcelists = [
                (s, Sourcelist.parse(s, context=self))
                for s in sections
                if s.id == "sourcelist"
            ]
            try:
                yield Sources(
                    tags,
                    list(zip(*sourcelists))[1] if sourcelists else [],
                    allow_duplicates,
                    default_to_implicit_numbering,
                    default_source_number_digits,
                )
            finally:
                for section, sourcelist in sourcelists:
                    section.data = sourcelist.get_raw_section_data()

    @ContextManager
    def patches(
        self,
        allow_duplicates: bool = False,
        default_to_implicit_numbering: bool = False,
        default_source_number_digits: int = 1,
    ) -> Generator[Patches, None, None]:
        """
        Context manager for accessing patches.

        Args:
            allow_duplicates: Whether to allow duplicate entries when adding new patches.
            default_to_implicit_numbering: Use implicit numbering (no source numbers) by default.
            default_source_number_digits: Default number of digits in a source number.

        Yields:
            Spec file patches as `Patches` object.
        """
        with self.sections() as sections, self.tags() as tags:
            patchlists = [
                (s, Sourcelist.parse(s, context=self))
                for s in sections
                if s.id == "patchlist"
            ]
            try:
                yield Patches(
                    tags,
                    list(zip(*patchlists))[1] if patchlists else [],
                    allow_duplicates,
                    default_to_implicit_numbering,
                    default_source_number_digits,
                )
            finally:
                for section, patchlist in patchlists:
                    section.data = patchlist.get_raw_section_data()

    @property
    def has_autochangelog(self) -> bool:
        """Whether the spec file uses %autochangelog."""
        with self.sections() as sections:
            try:
                section = sections.changelog
            except AttributeError:
                return False
            changelog = [ln.strip() for ln in section if ln.strip()]
            return changelog == ["%autochangelog"]

    def add_changelog_entry(
        self,
        entry: Union[str, List[str]],
        author: Optional[str] = None,
        email: Optional[str] = None,
        timestamp: Optional[Union[datetime.date, datetime.datetime]] = None,
        evr: Optional[str] = None,
    ) -> None:
        """
        Adds a new %changelog entry. Does nothing if there is no %changelog section
        or if %autochangelog is being used.

        If not specified, author and e-mail will be determined using rpmdev-packager, if available.
        Timestamp, if not set, will be set to current time (in local timezone).

        Args:
            entry: Entry text or list of entry lines.
            author: Author of the entry.
            email: E-mail of the author.
            timestamp: Timestamp of the entry.
              Supply `datetime` rather than `date` for extended format.
            evr: Override the EVR part of the changelog entry.
              Macros will be expanded automatically. By default, the function
              determines the appropriate value based on the specfile's current
              %{epoch}, %{version}, and %{release} values.
        """
        if self.has_autochangelog:
            return
        if evr is None:
            evr = "%{?epoch:%{epoch}:}%{version}-%{release}"
        with self.changelog() as changelog:
            if changelog is None:
                return
            evr = self.expand(evr, extra_macros=[("dist", "")])
            if isinstance(entry, str):
                entry = [entry]
            if timestamp is None:
                # honor the timestamp format, but default to date-only
                if changelog and changelog[-1].extended_timestamp:
                    timestamp = datetime.datetime.now().astimezone()
                else:
                    timestamp = datetime.date.today()
            if author is None:
                try:
                    author = subprocess.check_output("rpmdev-packager").decode().strip()
                except (FileNotFoundError, subprocess.CalledProcessError) as e:
                    raise SpecfileException("Failed to auto-detect author") from e
            elif email is not None:
                author += f" <{email}>"
            if changelog:
                # try to preserve padding of day of month
                padding = max(
                    (e.day_of_month_padding for e in reversed(changelog)), key=len
                )
            else:
                padding = "0"
            changelog.append(
                ChangelogEntry.assemble(
                    timestamp,
                    author,
                    entry,
                    evr,
                    day_of_month_padding=padding,
                    append_newline=bool(changelog),
                )
            )

    def _tag(name: str, doc: str) -> property:  # type: ignore[misc]
        """
        Returns a property that allows to get/set value of a specified tag.

        Args:
            name: Tag name.
            doc: Property docstring.

        Returns:
            Tag value property.
        """

        def getter(self) -> Optional[str]:
            with self.tags() as tags:
                try:
                    return getattr(tags, name).value
                except AttributeError:
                    return None

        def setter(self, value: str) -> None:
            with self.tags() as tags:
                getattr(tags, name).value = value

        return property(getter, setter, doc=doc)

    def _expanded_tag(name: str, doc: str) -> property:  # type: ignore[misc]
        """
        Returns a property that allows to get expanded value of a specified tag.

        Args:
            name: Tag name.
            doc: Property docstring.

        Returns:
            Expanded tag value property.
        """

        def getter(self) -> Optional[str]:
            with self.tags() as tags:
                try:
                    return getattr(tags, name).expanded_value
                except AttributeError:
                    return None

        return property(getter, doc=doc)

    name = _tag("name", "Name as stored in the spec file.")
    expanded_name = _expanded_tag("name", "Name with macros expanded.")

    version = _tag("version", "Version as stored in the spec file.")
    expanded_version = _expanded_tag("version", "Version with macros expanded.")

    raw_release = _tag("release", "Release string as stored in the spec file.")
    expanded_raw_release = _expanded_tag(
        "release", "Release string with macros expanded."
    )

    summary = _tag("summary", "Summary as stored in the spec file.")
    expanded_summary = _expanded_tag("summary", "Summary with macros expanded.")

    license = _tag("license", "License as stored in the spec file.")
    expanded_license = _expanded_tag("license", "License with macros expanded.")

    epoch = _tag("epoch", "Epoch as stored in the spec file.")
    expanded_epoch = _expanded_tag("epoch", "Epoch with macros expanded.")

    url = _tag("url", "URL as stored in the spec file.")
    expanded_url = _expanded_tag("url", "URL with macros expanded.")

    @staticmethod
    def _split_raw_release(
        raw_release: str,
    ) -> Tuple[str, Optional[str], Optional[int]]:
        """
        Splits raw release string into release, dist and minorbump parts.

        Args:
            raw_release: Raw release string.

        Returns:
            Tuple of (release, dist, minorbump).
        """
        tokens = re.split(r"(%(?P<m>\{\??)?dist(?(m)\}))(\.(\d+))?$", raw_release)
        if len(tokens) == 1:
            return tokens[0], None, None
        release, dist, _, _, minorbump, *_ = tokens
        return release, dist, int(minorbump) if minorbump else None

    @classmethod
    def _get_updated_release(cls, raw_release: str, release: str) -> str:
        """
        Returns the specified raw release string updated with the specified release.
        Minorbump, if there is one, is removed.

        Args:
            raw_release: Raw release string.
            release: New release.

        Returns:
            Updated raw release string.
        """
        dist = cls._split_raw_release(raw_release)[1] or ""
        return f"{release}{dist}"

    @property
    def release(self) -> str:
        """Release string without the dist suffix."""
        return self._split_raw_release(self.raw_release)[0]

    @release.setter
    def release(self, value: str) -> None:
        self.raw_release = self._get_updated_release(self.raw_release, value)

    @property
    def expanded_release(self) -> str:
        """Release string without the dist suffix with macros expanded."""
        return self.expand(self.release, extra_macros=[("dist", "")])

    def set_version_and_release(self, version: str, release: str = "1") -> None:
        """
        Sets both version and release at the same time.

        Args:
            version: Version string.
            release: Release string, defaults to '1'.
        """
        with self.tags() as tags:
            tags.version.value = version
            tags.release.value = self._get_updated_release(tags.release.value, release)

    def add_patch(
        self,
        location: str,
        number: Optional[int] = None,
        comment: Optional[str] = None,
        initial_number: int = 0,
        number_digits: int = 4,
    ) -> None:
        """
        Adds a patch.

        Args:
            location: Patch location (filename or URL).
            number: Patch number. It will be auto-assigned if not specified.
              If specified, it must be higher than any existing patch number.
            comment: Associated comment.
            initial_number: Auto-assigned number to start with if there are no patches.
            number_digits: Number of digits in the patch number.

        Raises:
            SourceNumberException when the specified patch number is not higher
              than any existing patch number.
        """
        with self.patches(default_source_number_digits=number_digits) as patches:
            highest_number = max((p.number for p in patches), default=-1)
            if number is not None:
                if number <= highest_number:
                    raise SourceNumberException(
                        "Patch number must be higher than any existing patch number"
                    )
            else:
                number = max(highest_number + 1, initial_number)
            index = patches.insert_numbered(number, location)
            if comment:
                patches[index].comments.extend(comment.splitlines())

    def update_value(
        self,
        value: str,
        requested_value: str,
        position: int,
        protected_entities: Optional[str] = None,
    ) -> str:
        """
        Updates a value from within the context of the spec file with a new value,
        but tries to preserve substitutions of locally defined macros and tags,
        updating the respective macro definitions and tag values instead.

        Args:
            value: Value to update.
            requested_value: Requested new value.
            position: Position (line number) of the value in the spec file.
            protected_entities: Regular expression specifying protected tags and macro definitions,
              ensuring their values won't be updated.

        Returns:
            Updated value. Can be equal to the original value.
        """

        @dataclass
        class Entity:
            name: str
            value: str
            type: Type
            position: int
            locked: bool = False
            updated: bool = False

        protected_regex = re.compile(
            # (?!) doesn't match anything
            protected_entities or "(?!)",
            re.IGNORECASE,
        )
        # collect modifiable entities
        entities = []
        with self.macro_definitions() as macro_definitions:
            entities.extend(
                [
                    Entity(
                        md.name, md.body, type(md), md.get_position(macro_definitions)
                    )
                    for md in macro_definitions
                    if not protected_regex.match(md.name)
                    and not md.name.endswith(")")  # skip macro definitions with options
                ]
            )
        with self.tags() as tags:
            entities.extend(
                [
                    Entity(t.name.lower(), t.value, type(t), t.get_position(tags))
                    for t in tags
                    if not protected_regex.match(t.name)
                ]
            )
        entities.sort(key=lambda e: e.position)

        def update(value, requested_value, position):
            modifiable_entities = {e.name for e in entities if e.position < position}
            # tags can be referenced as %{tag} or %{TAG}
            modifiable_entities.update(
                e.name.upper()
                for e in entities
                if e.position < position and e.type == Tag
            )
            regex, template = ValueParser.construct_regex(
                value, modifiable_entities, context=self
            )
            m = regex.match(requested_value)
            if m:
                d = m.groupdict()
                for grp, val in d.items():
                    if grp.startswith(SUBSTITUTION_GROUP_PREFIX):
                        continue
                    # find the closest matching entity
                    entity = [
                        e
                        for e in entities
                        if e.position < position
                        and (
                            e.name == grp
                            and e.type == MacroDefinition
                            or e.name == grp.lower()
                            and e.type == Tag
                        )
                    ][-1]
                    if entity.locked:
                        # avoid infinite recursion
                        return requested_value
                    entity.locked = True
                    try:
                        entity.value = update(entity.value, val, entity.position)
                    finally:
                        entity.locked = False
                        entity.updated = True
                return template.substitute(d)
            # no match, simply return the requested value
            return requested_value

        result = update(value, requested_value, position)
        # synchronize back any changes
        with self.macro_definitions() as macro_definitions:
            for entity in [
                e for e in entities if e.updated and e.type == MacroDefinition
            ]:
                getattr(macro_definitions, entity.name).body = entity.value
        with self.tags() as tags:
            for entity in [e for e in entities if e.updated and e.type == Tag]:
                getattr(tags, entity.name).value = entity.value
        return result

    def update_tag(
        self, name: str, value: str, protected_entities: str = ".*name"
    ) -> None:
        """
        Updates value of the given tag, trying to preserve substitutions
        of locally defined macros and tags, updating the respective macro definitions
        and tag values instead.

        Args:
            name: Tag name.
            value: Requested new value.
            protected_entities: Regular expression specifying protected tags and macro definitions,
              ensuring their values won't be updated.
        """
        with self.tags() as tags:
            tag = getattr(tags, name)
            original_value = tag.value
            position = tag.get_position(tags)
        # we can't use update_value() within the context manager, because any changes
        # made by it to tags or macro definitions would be thrown away
        updated_value = self.update_value(
            original_value, value, position, protected_entities=protected_entities
        )
        with self.tags() as tags:
            getattr(tags, name).value = updated_value



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy

import pytest
from flexmock import flexmock

from specfile.options import Options, Token, TokenType
from specfile.sections import Section, Sections


def test_find():
    sections = Sections(
        [
            Section("package"),
            Section("prep"),
            Section("changelog"),
            Section("pAckage", Options([Token(TokenType.DEFAULT, "foo")]), " "),
        ]
    )
    assert sections.find("prep") == 1
    assert sections.find("package foo") == 3
    with pytest.raises(ValueError):
        sections.find("install")


def test_get():
    sections = Sections(
        [
            Section("package"),
            Section("package", Options([Token(TokenType.DEFAULT, "bar")]), " "),
        ]
    )
    assert sections.get("package") == []
    assert sections.get("package bar") == []
    with pytest.raises(ValueError):
        sections.get("package foo")


def test_parse():
    sections = Sections.parse(
        [
            "0",
            "",
            "",
            "%description  %{desc}",
            "",
            "%prep",
            "0",
            "1",
            "2",
            "",
            "%description x %{desc}",
            "",
            "%package %{subpkg}",
            "",
            "%package x",
            "%files y",
            "0",
            "%changelog",
        ],
        context=flexmock(expand=lambda s, **_: "\n" if s == "%{desc}" else ""),
    )
    assert sections[0][0] == "0"
    assert sections[1].id == "description"
    assert sections.description == ["%{desc}", ""]
    assert sections[2].id == "prep"
    assert sections.prep == ["0", "1", "2", ""]
    assert sections[3].id == "description x"
    assert sections[3] == ["%{desc}", ""]
    assert sections[4].id == "package %{subpkg}"
    assert sections[4] == [""]
    assert sections[5].id == "package x"
    assert not sections[5]
    assert sections[-1].id == "changelog"


def test_parse_case_insensitive():
    sections = Sections.parse(
        ["0", "", "%Prep", "0", "1", "2", "", "%pAckage x", "Requires: bar"]
    )
    assert sections[0][0] == "0"
    assert sections[1].id == "Prep"
    assert sections.prep == ["0", "1", "2", ""]
    assert sections[2].id == "pAckage x"
    assert sections[2] == ["Requires: bar"]


def test_parse_invalid_name():
    sections = Sections.parse(
        [
            "%description",
            "This is a description.",
            "",
            "%description(fr)",
            "Ceci est une description.",
            "",
        ]
    )
    assert len(sections) == 2  # including empty preamble
    assert sections[1].id == "description"
    assert sections.description[2] == "%description(fr)"


def test_parse_macro_definitions():
    sections = Sections.parse(
        [
            "%package -n test",
            "Summary: Subpackage test",
            "",
            "%description -n test",
            "Subpackage test.",
            "",
            "%define template1()\\",
            "%package -n %{1}\\",
            "Summary: Subpackage %{1}\\",
            "\\",
            "%description -n %{1}\\",
            "Subpackage %{1}.",
            "",
            "%define template2() %{expand:",
            "%package -n %{1}",
            "Summary: Subpackage %{1}",
            "",
            "%description -n %{1}",
            "Subpackage %{1}.}",
            "",
            "%prep",
            "%autosetup",
        ]
    )
    assert len(sections) == 4
    assert sections[1].id == "package -n test"
    assert sections[-1].id == "prep"


def test_get_raw_data():
    sections = Sections(
        [
            Section("package", data=["0", "", ""]),
            Section("description", separator="  ", data=["%{desc}", ""]),
            Section("prep", data=["0", "1", "2", ""]),
            Section(
                "description",
                Options([Token(TokenType.DEFAULT, "x")]),
                " ",
                " ",
                ["%{desc}", ""],
            ),
            Section(
                "package",
                Options([Token(TokenType.DEFAULT, "%{subpkg}")]),
                " ",
                data=[""],
            ),
            Section("package", Options([Token(TokenType.DEFAULT, "x")]), " "),
            Section("files", Options([Token(TokenType.DEFAULT, "y")]), " ", data=["0"]),
            Section("changelog"),
        ],
    )
    assert sections.get_raw_data() == [
        "0",
        "",
        "",
        "%description  %{desc}",
        "",
        "%prep",
        "0",
        "1",
        "2",
        "",
        "%description x %{desc}",
        "",
        "%package %{subpkg}",
        "",
        "%package x",
        "%files y",
        "0",
        "%changelog",
    ]


def test_copy_sections():
    sections = Sections([Section("package", data=["Name: test", "Version: 0.1"])])
    shallow_copy = copy.copy(sections)
    assert shallow_copy == sections
    assert shallow_copy is not sections
    assert shallow_copy[0] is sections[0]
    deep_copy = copy.deepcopy(sections)
    assert deep_copy == sections
    assert deep_copy is not sections
    assert deep_copy[0] is not sections[0]


@pytest.mark.parametrize(
    "section, is_script",
    [
        ("package", False),
        ("install", True),
        ("PostUn", True),
        ("SourceList", False),
    ],
)
def test_is_script(section, is_script):
    assert Section(section).is_script == is_script



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import collections
import contextlib
import functools
import io
import os
import pickle
import sys
import tempfile
import types
from typing import Any, Callable, Dict, Generator, List, Optional, overload


@contextlib.contextmanager
def capture_stderr() -> Generator[List[bytes], None, None]:
    """
    Context manager for capturing output to stderr. A stderr output of anything run
    in its context will be captured in the target variable of the with statement.

    Yields:
        List of captured lines.
    """
    fileno = sys.__stderr__.fileno()
    with tempfile.TemporaryFile() as stderr, os.fdopen(os.dup(fileno)) as backup:
        sys.stderr.flush()
        os.dup2(stderr.fileno(), fileno)
        data: List[bytes] = []
        try:
            yield data
        finally:
            sys.stderr.flush()
            os.dup2(backup.fileno(), fileno)
            stderr.flush()
            stderr.seek(0, io.SEEK_SET)
            data.extend(stderr.readlines())


class GeneratorContextManager(contextlib._GeneratorContextManager):
    """
    Extended contextlib._GeneratorContextManager that provides get() method.
    """

    def __init__(self, function: Callable) -> None:
        super().__init__(function, tuple(), {})

    def __del__(self) -> None:
        # make sure the generator is fully consumed, as it is possible
        # that neither __enter__() nor content() have been called
        collections.deque(self.gen, maxlen=0)

    @property
    def content(self) -> Any:
        """
        Fully consumes the underlying generator and returns the yielded value.

        Returns:
            Value that would normally be the target variable of an associated with statement.

        Raises:
            StopIteration if the underlying generator is already exhausted.
        """
        result = next(self.gen)
        next(self.gen, None)
        return result


class ContextManager:
    """
    Class for decorating generator functions that should act as a context manager.

    Just like with contextlib.contextmanager, the generator returned from the decorated function
    must yield exactly one value that will be used as the target variable of the with statement.
    If the same function with the same arguments is called again from within previously generated
    context, the generator will be ignored and the target variable will be reused.

    Attributes:
        function: Decorated generator function.
        generators: Mapping of serialized function arguments to generators.
        values: Mapping of serialized function arguments to yielded values.
    """

    def __init__(self, function: Callable) -> None:
        self.function = function
        self.is_bound = False
        self.generators: Dict[bytes, Generator[Any, None, None]] = {}
        self.values: Dict[bytes, Any] = {}
        functools.update_wrapper(self, function)

    @overload
    def __get__(self, obj: None, objtype: Optional[type] = None) -> "ContextManager":
        pass

    @overload
    def __get__(self, obj: object, objtype: Optional[type] = None) -> types.MethodType:
        pass

    # implementing __get__() makes the class a non-data descriptor,
    # so it can be used as method decorator
    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        self.is_bound = True
        return types.MethodType(self, obj)

    def __call__(self, *args: Any, **kwargs: Any) -> GeneratorContextManager:
        # serialize the passed arguments
        payload = list(args) + sorted(kwargs.items())
        if payload and self.is_bound:
            # do not attempt to pickle self/cls
            payload[0] = (type(payload[0]), id(payload[0]))
        key = pickle.dumps(payload, protocol=pickle.HIGHEST_PROTOCOL)
        if (
            key in self.generators
            # gi_frame is None only in case generator is exhausted
            and self.generators[key].gi_frame is not None  # type: ignore[attr-defined]
        ):
            # generator is suspended, use existing value
            def existing_value():
                try:
                    yield self.values[key]
                except KeyError:
                    # if the generator is being consumed in GeneratorContextManager destructor,
                    # self.values[key] could have already been deleted
                    pass

            return GeneratorContextManager(existing_value)
        # create the generator
        self.generators[key] = self.function(*args, **kwargs)
        # first iteration yields the value
        self.values[key] = next(self.generators[key])

        def new_value():
            try:
                yield self.values[key]
            finally:
                # second iteration wraps things up
                next(self.generators[key], None)
                # the generator is now exhausted and the value is no longer valid
                del self.generators[key]
                del self.values[key]

        return GeneratorContextManager(new_value)



#!/usr/bin/env python3

# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import re
import sys
from pathlib import Path

import click
from extract import extract_arches, extract_sections, extract_tags

SECTIONS_SOURCE = Path("build/parseSpec.c")
TAGS_SOURCE = Path("build/parsePreamble.c")
ARCHES_SOURCE = Path("rpmrc.in")


@click.command(
    help="Update CONSTANTS_FILE with constants extracted from RPM_SOURCE_TREE."
)
@click.argument(
    "constants_file",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    required=True,
    default=Path("specfile/constants.py"),
)
@click.argument(
    "rpm_source_tree",
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    required=True,
)
def update_constants(constants_file: Path, rpm_source_tree: Path) -> None:
    section_names = extract_sections(rpm_source_tree / SECTIONS_SOURCE)
    tag_names = extract_tags(rpm_source_tree / TAGS_SOURCE)
    tags_with_args = extract_tags(rpm_source_tree / TAGS_SOURCE, with_args=True)
    arch_names = extract_arches(rpm_source_tree / ARCHES_SOURCE)
    content = original_content = constants_file.read_text()
    for constant, values in (
        ("SECTION_NAMES", section_names),
        ("TAG_NAMES", tag_names),
        ("TAGS_WITH_ARGS", tags_with_args),
        ("ARCH_NAMES", arch_names),
    ):
        formatted_values = "\n".join([f'    "{v}",' for v in sorted(set(values))])
        content = re.sub(
            rf"({constant}\s*=\s*{{).*?(}})",
            rf"\g<1>\n{formatted_values}\n\g<2>",
            content,
            flags=re.DOTALL,
        )
    if content == original_content:
        sys.exit(100)
        return
    constants_file.write_text(content)


if __name__ == "__main__":
    update_constants()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from specfile.options import Options, Positionals, Token, TokenType


@pytest.mark.parametrize(
    "optstring, tokens, result",
    [
        (
            "vp:m:M:",
            [
                Token(TokenType.DEFAULT, "28"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DOUBLE_QUOTED, "test arg"),
            ],
            [0, 2],
        ),
        (
            "vp:m:M:",
            [
                Token(TokenType.DEFAULT, "-p"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "1"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-v"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "28"),
            ],
            [6],
        ),
        (
            "vp:m:M:",
            [
                Token(TokenType.DEFAULT, "-m"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "10"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "test"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-M"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "20"),
            ],
            [4],
        ),
    ],
)
def test_positionals_get_items(optstring, tokens, result):
    options = Options(tokens, optstring)
    assert Positionals(options)._get_items() == result


@pytest.mark.parametrize(
    "optstring, tokens, index, value, tokens_index, token_type",
    [
        (
            "vp:m:M:",
            [
                Token(TokenType.DEFAULT, "-p1"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "test"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "28"),
            ],
            0,
            "test arg",
            2,
            TokenType.DOUBLE_QUOTED,
        ),
        (
            "vp:m:M:",
            [
                Token(TokenType.DEFAULT, "-p1"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "test"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "28"),
            ],
            1,
            123,
            4,
            TokenType.DEFAULT,
        ),
        (
            "vp:m:M:",
            [],
            0,
            "test",
            0,
            TokenType.DEFAULT,
        ),
    ],
)
def test_positionals_insert(optstring, tokens, index, value, tokens_index, token_type):
    options = Options(tokens, optstring)
    positionals = Positionals(options)
    positionals.insert(index, value)
    assert options._tokens[tokens_index].type == token_type
    assert options._tokens[tokens_index].value == str(value)


@pytest.mark.parametrize(
    "optstring, option, valid",
    [
        ("a:b:cDn:Tq", "a", True),
        ("a:b:cDn:Tq", "q", True),
        ("a:b:cDn:Tq", "v", False),
    ],
)
def test_options_valid_option(optstring, option, valid):
    options = Options([], optstring)
    assert options._valid_option(option) == valid


@pytest.mark.parametrize(
    "optstring, option, requires_argument",
    [
        ("a:b:cDn:Tq", "a", True),
        ("a:b:cDn:Tq", "q", False),
        ("a:b:cDn:Tq", "v", None),
    ],
)
def test_options_requires_argument(optstring, option, requires_argument):
    options = Options([], optstring)
    if option in optstring:
        assert options._requires_argument(option) == requires_argument
    else:
        with pytest.raises(ValueError):
            options._requires_argument(option)


@pytest.mark.parametrize(
    "optstring, tokens, option, result",
    [
        (
            "P:p:REb:z:F:d:o:Z",
            [
                Token(TokenType.DEFAULT, "-p1"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-b"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, ".test"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-E"),
            ],
            "p",
            (0, 0),
        ),
        (
            "P:p:REb:z:F:d:o:Z",
            [
                Token(TokenType.DEFAULT, "-p1"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-b"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, ".test"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-E"),
            ],
            "b",
            (2, 4),
        ),
        (
            "P:p:REb:z:F:d:o:Z",
            [
                Token(TokenType.DEFAULT, "-p1"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-b"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, ".test"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-E"),
            ],
            "E",
            (6, None),
        ),
        (
            "P:p:REb:z:F:d:o:Z",
            [
                Token(TokenType.DEFAULT, "-p1"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-b"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, ".test"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-E"),
            ],
            "F",
            (None, None),
        ),
    ],
)
def test_options_find_option(optstring, tokens, option, result):
    options = Options(tokens, optstring)
    assert options._find_option(option) == result


@pytest.mark.parametrize(
    "option_string, result",
    [
        (
            "-p1 -b .test -E",
            [
                Token(TokenType.DEFAULT, "-p1"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-b"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, ".test"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-E"),
            ],
        ),
        (
            "-p 28 -b .test\\ escape",
            [
                Token(TokenType.DEFAULT, "-p"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "28"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-b"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, ".test escape"),
            ],
        ),
        (
            '-b ".test \\"double quotes\\""',
            [
                Token(TokenType.DEFAULT, "-b"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DOUBLE_QUOTED, '.test "double quotes"'),
            ],
        ),
        (
            "-p1 -b .test_whitespace_at_the_end -M 2  ",
            [
                Token(TokenType.DEFAULT, "-p1"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-b"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, ".test_whitespace_at_the_end"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "-M"),
                Token(TokenType.WHITESPACE, " "),
                Token(TokenType.DEFAULT, "2"),
                Token(TokenType.WHITESPACE, "  "),
            ],
        ),
    ],
)
def test_options_tokenize(option_string, result):
    assert Options.tokenize(option_string) == result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock

from specfile.value_parser import (
    BuiltinMacro,
    ConditionalMacroExpansion,
    EnclosedMacroSubstitution,
    Macros,
    MacroSubstitution,
    ShellExpansion,
    StringLiteral,
    ValueParser,
)


@pytest.mark.parametrize(
    "value, nodes",
    [
        ("", []),
        ("escaped %%macro", [StringLiteral("escaped %%macro")]),
        (
            "%(echo %{version} | cut -d. -f3)",
            [ShellExpansion("echo %{version} | cut -d. -f3")],
        ),
        ("%version", [MacroSubstitution("version")]),
        ("1%{?dist}", [StringLiteral("1"), EnclosedMacroSubstitution("?dist")]),
        ("%{longdesc %{name}}", [EnclosedMacroSubstitution("longdesc %{name}")]),
        (
            "%{fedorarel}%{?dist}.2",
            [
                EnclosedMacroSubstitution("fedorarel"),
                EnclosedMacroSubstitution("?dist"),
                StringLiteral(".2"),
            ],
        ),
        (
            "%{?prever:0.}%{mainrel}%{?prever:.%{prerpmver}}",
            [
                ConditionalMacroExpansion("?prever", [StringLiteral("0.")]),
                EnclosedMacroSubstitution("mainrel"),
                ConditionalMacroExpansion(
                    "?prever",
                    [StringLiteral("."), EnclosedMacroSubstitution("prerpmver")],
                ),
            ],
        ),
        (
            '%{lua:ver = string.gsub(rpm.expand("%{ver}"), "-", "~"); print(string.lower(ver))}',
            [
                BuiltinMacro(
                    "lua",
                    'ver = string.gsub(rpm.expand("%{ver}"), "-", "~"); print(string.lower(ver))',
                )
            ],
        ),
    ],
)
def test_parse(value, nodes):
    assert ValueParser.parse(value) == nodes


@pytest.mark.parametrize(
    "value, macros, entries, regex, template",
    [
        (
            "%(echo %version | cut -d. -f1,2).0",
            {"%(echo %version | cut -d. -f1,2)": "1.1"},
            [],
            "^1\\.1(?P<sub_0>.+?)$",
            "%(echo %version | cut -d. -f1,2)${sub_0}",
        ),
        (
            "%version",
            {"%version": "1.0"},
            ["version"],
            "^(?P<version>.+?)$",
            "%version",
        ),
        (
            "%?version",
            {"%?version": "1.0"},
            ["version"],
            "^(?P<version>.+?)$",
            "%?version",
        ),
        (
            "%!?version",
            {"%!?version": ""},
            ["version"],
            "^$",
            "%!?version",
        ),
        (
            "1%{?dist}",
            {"%{?dist}": ".fc35"},
            [],
            "^(?P<sub_0>.+?)\\.fc35$",
            "${sub_0}%{?dist}",
        ),
        (
            "%{longdesc %name}",
            {"%{longdesc %name}": "Lorem ipsum dolor sit amet"},
            ["longdesc"],
            "^Lorem\\ ipsum\\ dolor\\ sit\\ amet$",
            "%{longdesc %name}",
        ),
        (
            "%{?prever:0.}%{mainrel}%{?prever:.%{prerpmver}}",
            {
                "%?prever": "beta3",
                "%{prerpmver}": "%(echo \"%{?prever}\" | sed -e 's|-||g')",
                "%{mainrel}": "2",
            },
            ["prever", "prerpmver", "mainrel"],
            "^(?P<sub_0>.+?)\\.(?P<mainrel>.+?)\\.(?P<prerpmver>.+?)$",
            "%{?prever:${sub_0}.}%{mainrel}%{?prever:.%{prerpmver}}",
        ),
        (
            "%{mingw64_libdir}/lib%{pkgname}stub%{majorver1}%{majorver2}.a",
            {
                "%{mingw64_libdir}": "/usr/x86_64-w64-mingw32/sys-root/mingw/lib",
                "%{pkgname}": "tk",
                "%{majorver1}": "8",
                "%{majorver2}": "6",
            },
            ["pkgname", "majorver1", "majorver2"],
            "^$",
            "%{mingw64_libdir}/lib%{pkgname}stub%{majorver1}%{majorver2}.a",
        ),
    ],
)
def test_construct_regex(value, macros, entries, regex, template):
    flexmock(Macros, expand=lambda m: macros.get(m, ""))
    r, t = ValueParser.construct_regex(value, entries)
    assert r.pattern == regex
    assert t.template == template
